import numpy as np
from pandas.core.frame import DataFrame
from akerbp.models.rule_based_models import BS_helpers
from akerbp.models.rule_based_models import crossplot_helpers


def flag_dencorr(
    df_well: DataFrame, y_pred: DataFrame = None, **kwargs
) -> DataFrame:
    """
    Returns anomalous density corrections

    Args:
        df_well (DataFrame): data from one well to analyse

    Returns:
        DataFrame: dataframe with added columns with results
    """
    print("Method: Density correction...")
    if y_pred is None:
        y_pred = df_well.copy()
    x = "DEN"
    y = "DENC"
    df_well = BS_helpers.find_BS_jumps(df_well).copy()
    dencorr_anomalies = []
    for bsr in df_well["BS_region"].unique():
        df_gr = df_well[df_well["BS_region"] == bsr].copy()
        unacceptable_denc = set(df_gr.loc[np.abs(df_gr["DENC"] > 0.1)].index.tolist())
        if (df_gr[x].dropna().shape[0] != 0) and (df_gr[y].dropna().shape[0] != 0):
            anomalies, scores, idx = crossplot_helpers.find_crossplot_scores(
                df_gr, x=x, y=y, **kwargs
            )
            # post process the results
            dencorr_anomalies.extend(
                list(set(anomalies).intersection(unacceptable_denc))
            )
            # scores
            for a_method, sc in scores.items():
                y_pred.loc[idx, f"{a_method}_dencorr"] = sc
                y_pred[f"{a_method}_dencorr"].fillna(0, inplace=True)
                # post process the results
                y_pred.loc[unacceptable_denc, f"{a_method}_dencorr"] = 1
    y_pred.loc[:, ["flag_dencorr_gen", "flag_dencorr_den"]] = 0, 0
    y_pred.loc[dencorr_anomalies, ["flag_dencorr_gen", "flag_dencorr_den"]] = 1
    return y_pred



from typing import Any, List, Dict
from pandas.core.frame import DataFrame
import os
import joblib
import pandas as pd
import importlib

from akerbp.mlpet import Dataset
from akerbp.models import Model
from akerbp.models import classification_models
from akerbp.models.rule_based_models import helpers


class BadlogModel(Model):
    """
    Badlog model class
    """

    def __init__(self, **kwargs):
        """
        Initializes badlog model
        """
        self.kwargs = kwargs
        for key, val in kwargs.items():
            setattr(self, key, val)

    def _validate_input(self, X: Any, **kwargs):
        """
        Validated input by checking type of X

        Args:
            X (DataFrame or dict): input data to model

        Raises:
            ValueError: if X is a dict and no metadata key
            ValueError: if X is a dict and no data key
            ValueError: if X not dict nor DataFrame
        """
        if isinstance(X, dict):
            if "metadata" not in X.keys():
                raise ValueError("'metadata' key is required")
            if "data" not in X.keys():
                raise ValueError("'data' key is required")
            X_df = pd.DataFrame.from_dict(X["data"])
            metadata = X["metadata"]
            self.X_pred, self.num_cols, self.cat_cols = helpers._apply_metadata(
                X_df, **kwargs.get("_metadata", metadata)
            )
        elif isinstance(X, pd.core.frame.DataFrame):
            X_df = X.copy()
            self.X_pred, self.num_cols, self.cat_cols = helpers._apply_metadata(
                X_df, **kwargs.get("_metadata", {})
            )
        else:
            raise ValueError("Please pass the data as a dict or a pandas DataFrame")

    def _init_cp_models(
        self,
        cp_names: List[str],
        folder_path: str,
        data_settings: Dict,
        model_settings: Dict,
        mappings: Dict,
    ):
        """
        Instantiate datasets and models for each of the crossplots

        Args:
            cp_names (list): list with crossplot names
            folder_path (str): path to folder with models
            data_settings (dict): settings of preprocessing for using the model
            model_settings (dict): settings of model XGBoost
            mappings (dict): mappings dictionary of cat variables

        Returns:
            tuple: dictionaries with dataset objects, models and key wells
        """
        datasets = dict()
        models = dict()
        key_wells = dict()
        folder_paths = dict()
        for cp_name in cp_names:
            print("init ds and model - BadlogModel - {}".format(cp_name))
            folder_paths[cp_name] = os.path.join(folder_path, cp_name)
            # Create dataset but don't load any data into the dataset, because
            # we need to get that from running the find_crossplot_scores
            datasets[cp_name] = Dataset(
                settings=data_settings,
                mappings=mappings,
                folder_path=folder_paths[cp_name],
            )

            models[cp_name] = classification_models.XGBoostClassificationModel(
                model_settings, folder_paths[cp_name]
            )
            key_wells[cp_name] = joblib.load(
                os.path.join(folder_paths[cp_name], "key_wells.joblib")
            )
        return datasets, models, key_wells

    def _get_cp_params(self, **kwargs):
        """
        Returns necessary variables to generate crossplots predictions

        Returns:
            tuple: path to models folder, settings for data and models,
            mappings dictionary
        """
        settings = kwargs.get("settings", None)
        mappings = kwargs.get("mappings", None)
        folder_path = kwargs.get("folder_path", None)
        # Assuming all cp_names use the same model- and data-settings
        if all([settings, mappings, folder_path]):
            model_settings = settings["models"]
            data_settings = settings["datasets"]
        else:
            raise ValueError(
                "Please provide settings, mappings and folder path for the crossplot method."
            )
        return folder_path, data_settings, model_settings, mappings

    def predict(self, X: Any, methods: List[str], **kwargs):
        """
        Args:
            X (dict or DataFrame): expected features
            methods (list): list of methods to apply on X

        Returns:
            y (DataFrame): including different badlog flags and scores
        """
        self._validate_input(X, **kwargs)
        # Validate feature
        self.X_pred = helpers._create_features(self.X_pred)
        self.X_pred_features = set(self.X_pred.columns)
        self.expected_curves = set(
            [
                "DEPTH",
                "DEN",
                "DENC",
                "AC",
                "ACS",
                "BS",
                "CALI",
                "GR",
                "NEU",
                "RDEP",
                "RMED",
                "RMIC",
                "PEF",
                "GROUP",
            ]
        )
        helpers._validate_features(self.X_pred_features, self.expected_curves)
        self.methods = methods
        self.y = pd.DataFrame()
        for method in self.methods:
            # Evaluate method
            method_helpers = importlib.import_module(
                f"akerbp.models.rule_based_models.{method}_helpers"
            )
            method_evaluator = getattr(method_helpers, f"flag_{method}")
            if method == "crossplot":
                cp_names = kwargs.get("cp_names", ["vp_den", "vp_vs", "ai_vpvs"])
                datasets, models, key_wells = self._init_cp_models(
                    cp_names, *self._get_cp_params(**kwargs)
                )
                method_flags = method_evaluator(
                    self.X_pred, cp_names, datasets, models, key_wells
                )
            else:
                method_flags = method_evaluator(self.X_pred)
            self.y[method_flags.columns] = method_flags
        result_columns = [
            col
            for col in self.y.columns
            if col.split("_")[0] in ["flag", "agg", "score"]
        ]
        return self.y[result_columns]



from typing import List, Tuple, Any
from pandas.core.frame import DataFrame
import numpy as np
import pandas as pd
from sklearn.covariance import EllipticEnvelope
from sklearn.svm import OneClassSVM
from sklearn.ensemble import IsolationForest
from pyod.utils.utility import standardizer
from akerbp.mlpet import Dataset


def get_params(**kwargs) -> Tuple:
    """
    Gets parameters for the outlier detection algorithms

    Returns:
        tuple: dictionaries with parameters per method
    """
    SVM_params = {"nu": kwargs.get("SVM_nu", 0.01)}
    IsoFor_params = {
        "n_estimators": kwargs.get("IsoFor_n_estimators", 50),
        "contamination": kwargs.get("IsoFor_contamination", 0.01),
        "random_state": kwargs.get("IsoFor_random_state", 0),
    }
    Abod_parameters = {"contamination": 0.1, "n_neighbors": 10}
    EliEnv_parameters = {"contamination": 0.05, "random_state": 0}
    Abod_parameters = {"contamination": 0.1, "n_neighbors": 10}
    EliEnv_parameters = {"contamination": 0.05, "random_state": 0}
    return SVM_params, IsoFor_params, Abod_parameters, EliEnv_parameters


def find_crossplot_scores(
    df: DataFrame, x: str, y: str, algo: str = "comb", **kwargs
) -> List:
    """
    Find anomalies based on the crossplots of x vs. y

    Args:
        df (DataFrame): dataframe with data from one well
        x (str): column for axis x in crossplot
        y (str): column for axis y in crossplot
        algo (str, optional): which algorithm to use. Defaults to 'comb'

    Returns:
        list: list with indices of anomalous samples
    """
    algo = algo.lower()
    df = df[(df[x] != 0) & (df[y] != 0)].copy()
    tmp = df[[x, y]].dropna().copy()
    anomal_idx = []
    method_scores = dict()
    return_scores = dict()
    SVM_params, IsoFor_params, _, EliEnv_parameters = get_params(**kwargs)
    # if there are not even 10 samples we return
    if len(tmp) < 10:
        return [], {}, tmp.index
    if (algo == "comb") or (algo == "elienv"):
        try:
            elienv = EllipticEnvelope(
                contamination=EliEnv_parameters["contamination"],
                random_state=EliEnv_parameters["random_state"],
            ).fit(tmp[[x, y]])
            preds2 = elienv.predict(tmp[[x, y]])
            original_scores = elienv.score_samples(tmp[[x, y]])
        except Exception as e:
            # FIXME: elliptic envelope throws covariance error when not so many
            # samples. # Quick fix for now.
            preds2 = np.zeros(len(tmp))
            original_scores = np.zeros(len(tmp))
            print(e)
        if algo == "elienv":
            preds = preds2
        method_scores["elienv"] = np.abs(standardizer(original_scores.reshape(-1, 1)))

    if (algo == "comb") or (algo == "svm"):
        svm = OneClassSVM(kernel="rbf", gamma="scale", nu=SVM_params["nu"]).fit(
            tmp[[x, y]]
        )
        preds3 = svm.predict(tmp[[x, y]])
        if algo == "svm":
            preds = preds3
        original_scores = svm.score_samples(tmp[[x, y]])
        method_scores["SVM"] = np.abs(standardizer(original_scores.reshape(-1, 1)))

    if (algo == "comb") or (algo == "isofor"):
        ifo = IsolationForest(
            n_estimators=IsoFor_params["n_estimators"],
            contamination=IsoFor_params["contamination"],
            random_state=IsoFor_params["random_state"],
        ).fit(tmp[[x, y]])
        preds4 = ifo.predict(tmp[[x, y]])
        if algo == "isofor":
            preds = preds4
        original_scores = ifo.score_samples(tmp[[x, y]])
        method_scores["IsoFor"] = standardizer(original_scores.reshape(-1, 1))

    if algo == "comb":
        preds = (
            np.where(preds2 == -1, 1, 0)
            + np.where(preds3 == -1, 1, 0)
            + np.where(preds4 == -1, 1, 0)
        )
        tmp["pred"] = np.where(preds > 1, 1, 0)
    else:
        tmp["pred"] = np.where(preds == -1, 1, 0)

    for k, v in method_scores.items():
        return_scores[k] = v[:, 0]

    return_scores = pd.DataFrame(return_scores)
    return_scores["agg"] = return_scores.mean(axis=1)
    anomal_idx.append(list(tmp[tmp.pred == 1].index))
    return sum(anomal_idx, []), return_scores, tmp.index


def train_crossplot(**kwargs):
    pass


def get_proba_predictions(
    X: DataFrame, y: DataFrame, bad_logs: DataFrame, model: Any, cp_name: str
) -> DataFrame:
    """
    Add to the dataframe the score results for each of the bad anomalies.

    Args:
        X (DataFrame): dataframe for prediction on
        y (DataFrame): datafrrame with columns to populate results
        bad_logs (DataFrame): bad log samples dataframe
        model (Any): trained model with Scikit learn or XGBoost model
        cp_name (str): crossplot name

    Returns:
        y: DataFrame populated with results
    """
    cp_preds = model.predict_proba(X)
    if cp_name == "vp_den":
        y.loc[bad_logs.index, "score_vpden_ac"] = cp_preds[:, 0]
        y.loc[bad_logs.index, "score_vpden_den"] = cp_preds[:, 1]
    elif cp_name == "vp_vs":
        y.loc[bad_logs.index, "score_vpvs_ac"] = cp_preds[:, 0]
        y.loc[bad_logs.index, "score_vpvs_acs"] = cp_preds[:, 1]
    elif cp_name == "ai_vpvs":
        y.loc[bad_logs.index, "score_aivpvs_ac"] = cp_preds[:, 0]
        y.loc[bad_logs.index, "score_aivpvs_acs"] = cp_preds[:, 1]
        y.loc[bad_logs.index, "score_aivpvs_den"] = cp_preds[:, 2]
    return y


def get_class_predictions(
    X: DataFrame, y: DataFrame, bad_logs: DataFrame, model: Any, cp_name: str
) -> DataFrame:
    """
    Add to the dataframe the class results for each of the bad anomalies.

    Args:
        X (DataFrame): dataframe for prediction on
        y (DataFrame): datafrrame with columns to populate results
        bad_logs (DataFrame): bad log samples dataframe
        models (dict): dictionary with models per cp
        cp_name (str): crossplot name

    Returns:
        y: pd.DataFrame populated with results
    """
    cp_preds = model.predict(X)
    y["tmp"] = 0
    y.loc[bad_logs.index, "tmp"] = cp_preds
    # Map predictions depending on which crossplot they come from
    if cp_name == "vp_den":
        y.loc[y.tmp.isin([1, 3]), "flag_vpden_ac"] = 1
        y.loc[y.tmp.isin([2, 3]), "flag_vpden_den"] = 1
    elif cp_name == "vp_vs":
        y.loc[y.tmp.isin([1, 3]), "flag_vpvs_ac"] = 1
        y.loc[y.tmp.isin([2, 3]), "flag_vpvs_acs"] = 1
    elif cp_name == "ai_vpvs":
        y.loc[y.tmp.isin([4, 5, 6, 7]), "flag_aivpvs_den"] = 1
        y.loc[y.tmp.isin([1, 3, 5, 7]), "flag_aivpvs_ac"] = 1
        y.loc[y.tmp.isin([2, 4, 6, 7]), "flag_aivpvs_acs"] = 1
    return y


def flag_well_crossplots(
    X: DataFrame, x: str, y: str, algo: str, cp_name: str, **kwargs
) -> Tuple[DataFrame, List]:
    """
    Returns anomalous indices of crossplots between x and y.
    Used by other methods.

    Args:
        X (DataFrame): dataframe with data to flag
        x (str): colum for x-axis in crossplot
        y (str): column for y-axis in crossplot
        algo (str): comb or specic algorithm to run
        cp_name (str): crossplot name

    Returns:
        Tuple[DataFrame, List]: dataframe with flagged samples and list
        of anamalous indices
    """
    res = dict(
        zip(
            ["anomalies", "scores", "idx"],
            find_crossplot_scores(X, x, y, algo=algo, **kwargs),
        )
    )
    for comb_method, scs in res["scores"].items():
        method_col = f"{comb_method}_{cp_name}"
        X.loc[res["idx"], method_col] = scs.values
        X[method_col].fillna(0, inplace=True)
    return X, res["anomalies"]


def flag_crossplots(
    X: DataFrame, x: str, y: str, algo: str, cp_name: str, **kwargs
) -> Tuple[DataFrame, List]:
    """
    Returns anomalous indices of crossplots between x and y

    Args:
        X (DataFrame): dataframe with data to flag
        x (str): colum for x-axis in crossplot
        y (str): column for y-axis in crossplot
        algo (str): comb or specic algorithm to run
        cp_name (str): crossplot name

    Returns:
        Tuple[DataFrame, List]: dataframe with flagged samples and list
        of anamalous indices
    """
    anomalies = []
    for lsu in X["GROUP"].unique():
        lsu_data = X[X["GROUP"] == lsu]
        res = dict(
            zip(
                ["anomalies", "scores", "idx"],
                find_crossplot_scores(lsu_data, x, y, algo=algo, **kwargs),
            )
        )
        for comb_method, scs in res["scores"].items():
            method_col = f"{comb_method}_{cp_name}"
            X.loc[res["idx"], method_col] = scs.values
            X[method_col].fillna(0, inplace=True)
        anomalies.extend(res["anomalies"])
    return X, anomalies


def classify_badlog(
    y: DataFrame, cp_name: str, ds: Dataset, key_wells: dict, model: Any
) -> DataFrame:
    """
    Returns both probabilities and class predictions of anomalous indices

    Args:
        y (DataFrame): dataframe with data to flag
        cp_name (str): crossplot name
        ds (Dataset): dataset object from akerbp.mlpet
        key_wells (dict): dictionary with key wells required for 2-point
        shift normalization
        model (Any): model form Scikit learn or XGBoost

    Returns:
        DataFrame: dataframe with columns of results added
    """
    bad_logs = ds.df_original[ds.df_original["flag_crossplot_gen"] != 0]
    # if no badlogs, dont do anything
    if len(bad_logs) == 0:
        return y

    # update the default preprocessing pipeline with the specifics required by this model
    kwargs = {
        "normalize_curves": {"user_key_wells": key_wells},
    }
    preprocess_kwargs = {}
    for func in ds.default_preprocessing_workflow:
        if func in kwargs:
            preprocess_kwargs[func] = kwargs[func]
        else:
            preprocess_kwargs[func] = {}
    # Preprocess data
    df_preprocessed = ds.preprocess(ds.df_original.loc[bad_logs.index], **kwargs)
    feats = model.get_booster().feature_names
    # scores
    y_scores = get_proba_predictions(
        df_preprocessed[feats], y, bad_logs, model, cp_name
    )
    y[y_scores.columns] = y_scores
    # flags
    y_flags = get_class_predictions(df_preprocessed[feats], y, bad_logs, model, cp_name)
    y[y_flags.columns] = y_flags
    return y


def flag_crossplot(
    X_pred: DataFrame,
    cp_names: List,
    datasets: dict,
    models: dict,
    key_wells: dict,
    **kwargs,
) -> DataFrame:
    """
    Main function to flag pre defined crossplots

    Args:
        X_pred (DataFrame): dataframe with data from one well
        cp_names (List): list of crossplots to run
        datasets (dict): dictionary with Dataset obejcts for each crossplot model
        models (dict): dictionary with models for each crossplot
        key_wells (dict): dictionary with key wells for normalization process

    Returns:
        DataFrame: dataframe with added columns with flag results
    """
    print("Method: crossplot - general...")
    # FIXME! The shouldn't preferably be hard-coded.
    outlier_methods = ["EliEnv", "SVM", "IsoFor"]
    for cp_name in cp_names:
        for method in outlier_methods:
            X_pred[f"{method}_{cp_name}"] = 0.0

    # get anomalies per crossplot and their scores
    vp_den_anomalies, ai_vpvs_anomalies, vp_vs_anomalies = [], [], []
    X_pred, vp_den_anomalies = flag_crossplots(
        X_pred, "VP", "DEN", "comb", "vp_den", **kwargs
    )
    X_pred, ai_vpvs_anomalies = flag_crossplots(
        X_pred, "AI", "VPVS", "comb", "ai_vpvs", **kwargs
    )
    X_pred, vp_vs_anomalies = flag_crossplots(
        X_pred, "VP", "VS", "comb", "vp_vs", **kwargs
    )

    y = X_pred.copy()
    y.loc[
        :,
        [
            "flag_vpden_ac",
            "flag_vpvs_ac",
            "flag_aivpvs_ac",
            "flag_vpvs_acs",
            "flag_aivpvs_acs",
            "flag_vpden_den",
            "flag_aivpvs_den",
        ],
    ] = (0, 0, 0, 0, 0, 0, 0)

    pred_gen_cols = [
        "flag_crossplot_gen",
        "flag_vpden_gen",
        "flag_aivpvs_gen",
        "flag_vpvs_gen",
    ]
    y.loc[:, pred_gen_cols] = 0, 0, 0, 0
    y.loc[vp_den_anomalies, ["flag_crossplot_gen", "flag_vpden_gen"]] = 1
    y.loc[ai_vpvs_anomalies, ["flag_crossplot_gen", "flag_aivpvs_gen"]] = 1
    y.loc[vp_vs_anomalies, ["flag_crossplot_gen", "flag_vpvs_gen"]] = 1

    # Add the general crossplot flags to X_pred(to use as features in the model)
    X_pred[pred_gen_cols] = y[pred_gen_cols]

    # Run the anomalous samples through the relevant trained model and get predictions
    # Preprocessing using key_wells will need the well_name column,
    # but the value is unimportant
    X_pred["well_name"] = "dummy"
    for cp_name in cp_names:
        print(f"Method: crossplot - {cp_name}...")
        ds = datasets[cp_name]
        ds.save_df_to_cls(X_pred)
        y = classify_badlog(y, cp_name, ds, key_wells[cp_name], models[cp_name])
    y = y.fillna(0)
    return y



import numpy as np
from pandas.core.frame import DataFrame
from akerbp.models.rule_based_models import helpers
import akerbp.models.rule_based_models.crossplot_helpers as crossplot_helpers


def flag_outliers(
    df_well: DataFrame, y_pred: DataFrame = None, **kwargs
) -> DataFrame:
    """
    Returns anomalous indices for AC, ACS and DEN based on outliers of crossplots
    between AC, ACS and DEN and other curves

    Args:
        df_well (DataFrame): data from one well

    Returns:
        tuple: lists of anomalous indices of AC, ACS and DEN
    """
    print("Method: outliers...")
    if "CALI_BS" not in df_well.columns:
        df_well["CALI_BS"] = np.abs(df_well["CALI"] - df_well["BS"])
    if "RDEP_log" not in df_well.columns and "RDEP" in df_well.columns:
        df_well["RDEP_log"] = np.log10(df_well.RDEP)
    default_curves = ["AC", "ACS", "CALI", "GR", "NEU", "PEF", "RDEP_log", "CALI_BS"]
    curves = kwargs.get("curves", default_curves)

    if "GROUP" in curves:
        curves.remove("GROUP")

    algo_params = {
        "DBSCAN_eps": 0.4,
        "EliEnv_contamination": 0.01,
        "EliEnv_random_state": 0,
        "SVM_gamma": "scale",
        "SVM_nu": 0.01,
        "IsoFor_n_estimators": 50,
        "IsoFor_contamination": 0.005,
        "IsoFor_random_state": 0
    }
    if y_pred is None:
        y_pred = df_well.copy()

    ac_outliers_anomalies = []
    acs_outliers_anomalies = []
    den_outliers_anomalies = []

    logname_curves = {
        "AC": ([c for c in curves if c not in ["VP", "AC"]], ac_outliers_anomalies),
        "ACS": ([c for c in curves if c not in ["VS", "ACS"]], acs_outliers_anomalies),
        "DEN": ([c for c in curves if c != "DEN"], den_outliers_anomalies)
    }

    method = "outliers"
    for logname, (curves, anomalies) in logname_curves.items():
        for y in curves:
            flags, scores, idx = crossplot_helpers.find_crossplot_scores(
                df_well, x=logname, y=y, **algo_params
            )
            anomalies.extend(flags)
            for a_method in scores.keys():
                y_pred.loc[
                    idx, f"{a_method}_{method}_{logname.lower()}"
                ] = scores[a_method]
                y_pred[
                    f"{a_method}_{method}_{logname.lower()}"
                ].fillna(0, inplace=True)

    y_pred.loc[:, [
        "flag_outliers_gen",
        "flag_outliers_ac",
        "flag_outliers_acs",
        "flag_outliers_den"
    ]] = 0, 0, 0, 0
    y_pred.loc[ac_outliers_anomalies, "flag_outliers_ac"] = 1
    y_pred.loc[acs_outliers_anomalies, "flag_outliers_acs"] = 1
    y_pred.loc[den_outliers_anomalies, "flag_outliers_den"] = 1

    y_pred["flag_outliers_ac"] = helpers.fill_holes(y_pred, "flag_outliers_ac")
    y_pred["flag_outliers_acs"] = helpers.fill_holes(y_pred, "flag_outliers_acs")
    y_pred["flag_outliers_den"] = helpers.fill_holes(y_pred, "flag_outliers_den")

    y_pred.loc[(
        (y_pred.flag_outliers_ac == 1) | (y_pred.flag_outliers_acs == 1) |\
        (y_pred.flag_outliers_den == 1)
    ), "flag_outliers_gen"] = 1

    return y_pred



from pandas.core.frame import DataFrame
import numpy as np
from akerbp.models.rule_based_models import helpers
import akerbp.models.rule_based_models.crossplot_helpers as crossplot_helpers


def flag_logtrend(
    df_well: DataFrame, y_pred: DataFrame = None, **kwargs
) -> DataFrame:
    """
    Returns anomalous indices for AC, ACS and DEN based on log trends

    Args:
        df_well (DataFrame): data from one well
        y_pred (DataFrame): output data to attach results to
        score (bool): if returned values should be score or binary

    Returns:
        tuple: lists of anomalous indices of AC, ACS and DEN
    """
    print("Method: logtrend...")
    if "CALI_BS" not in df_well.columns:
        df_well["CALI_BS"] = np.abs(df_well["CALI"] - df_well["BS"])
    default_curves = ["CALI", "GR", "NEU", "CALI_BS"]
    curves_to_diff = kwargs.get("curves_to_diff", default_curves)
    if "GROUP" in curves_to_diff:
        curves_to_diff.remove("GROUP")

    algo_params = {
        "DBSCAN_eps": 0.1,
        "EliEnv_contamination": 0.01,
        "EliEnv_random_state": 0,
        "SVM_gamma": "scale",
        "SVM_nu": 0.005,
        "IsoFor_n_estimators": 50,
        "IsoFor_contamination": 0.005,
        "IsoFor_random_state": 0
    }

    # replace curves values by difference of consecutive samples
    tmp_curves_to_diff = curves_to_diff +\
        [c for c in ["AC", "ACS", "DEN"] if c not in curves_to_diff]
    df_well[tmp_curves_to_diff] = df_well[tmp_curves_to_diff].diff()

    if y_pred is None:
        y_pred = df_well.copy()

    ac_trends_anomalies = []
    acs_trends_anomalies = []
    den_trends_anomalies = []

    logname_curves = {
        "AC": (
            [c for c in curves_to_diff if c not in ["VP", "AC"]],
            ac_trends_anomalies
        ),
        "ACS": (
            [c for c in curves_to_diff if c not in ["VS", "ACS"]],
            acs_trends_anomalies
        ),
        "DEN": (
            [c for c in curves_to_diff if c != "DEN"],
            den_trends_anomalies
        )
    }

    method = "logtrend"
    for logname, (curves, anomalies) in logname_curves.items():
        for y in curves:
            flags, scores, idx = crossplot_helpers.find_crossplot_scores(
                df_well, x=logname, y=y, **algo_params
            )
            anomalies.extend(flags)
            for a_method in scores.keys():
                y_pred.loc[
                    idx, f"{a_method}_{method}_{logname.lower()}"
                ] = scores[a_method]
                y_pred[
                    f"{a_method}_{method}_{logname.lower()}"
                ].fillna(0, inplace=True)

    y_pred.loc[:, [
        "flag_logtrend_gen",
        "flag_logtrend_ac",
        "flag_logtrend_acs",
        "flag_logtrend_den"
    ]] = 0, 0, 0, 0
    y_pred.loc[ac_trends_anomalies, "flag_logtrend_ac"] = 1
    y_pred.loc[acs_trends_anomalies, "flag_logtrend_acs"] = 1
    y_pred.loc[den_trends_anomalies, "flag_logtrend_den"] = 1

    y_pred["flag_logtrend_ac"] = helpers.fill_holes(y_pred, "flag_logtrend_ac")
    y_pred["flag_logtrend_acs"] = helpers.fill_holes(y_pred, "flag_logtrend_acs")
    y_pred["flag_logtrend_den"] = helpers.fill_holes(y_pred, "flag_logtrend_den")

    y_pred.loc[(
        (y_pred.flag_logtrend_ac == 1) | (y_pred.flag_logtrend_acs == 1) |\
        (y_pred.flag_logtrend_den == 1)
    ), "flag_logtrend_gen"] = 1

    return y_pred



from typing import List
from pandas.core.frame import DataFrame
import numpy as np
import pandas as pd
import umap as UMAP
from sklearn.preprocessing import MinMaxScaler
from sklearn.svm import OneClassSVM
from sklearn.ensemble import IsolationForest


def umap3d(df_well: DataFrame, curves_umap: List) -> List:
    """
    Returns anomalous indices found by umap clustering.
    Not in use at the moment.

    Args:
        df_well (DataFrame): data from one well
        curves_umap (list): list of curves to cluster

    Returns:
        list: list of anomalous indices
    """
    umap = UMAP.UMAP(n_components=3)
    df_well_umap = umap.fit_transform(
        MinMaxScaler().fit_transform(df_well[curves_umap])
    )
    df_well_umap = pd.DataFrame(
        df_well_umap, index=df_well.index, columns=["first", "second", "third"]
    )
    tmp_X = df_well_umap[["first", "second", "third"]]
    preds3 = OneClassSVM(gamma="scale", nu=0.001).fit_predict(tmp_X)
    preds4 = IsolationForest(n_estimators=50, contamination=0.001).fit_predict(tmp_X)
    preds = np.where(preds3 == -1, 1, 0) + np.where(preds4 == -1, 1, 0)
    tmp_X["pred"] = np.where(preds > 0, True, False)
    return list(tmp_X[tmp_X.pred == True].index)


def flag_umap(
    df_well: DataFrame, y_pred: DataFrame = None, **kwargs
) -> DataFrame:
    """
    Flags umap anomalies for AC, ACS and DEN

    Args:
        df_well ([type]): [description]

    Returns:
        [type]: [description]
    """
    print("Method: UMAP...")

    if y_pred is None:
        y_pred = df_well.copy()

    y_pred.loc[:, [
        "flag_umap_gen",
        "flag_umap_ac",
        "flag_umap_acs",
        "flag_umap_den"
    ]] = 0, 0, 0, 0

    curves_umap = ["AC", "GR", "NEU", "RDEP", "RMED", "VP", "AI", "VPVS"]
    try:
        ac_umap = umap3d(df_well, curves_umap)
    except:
        ac_umap = []
    curves_umap = ["ACS", "GR", "NEU", "RDEP", "RMED", "VS", "AI", "VPVS"]
    try:
        acs_umap = umap3d(df_well, curves_umap)
    except:
        acs_umap = []
    curves_umap = ["GR", "NEU", "RDEP", "RMED", "AI", "VPVS", "DEN", "DENC"]
    try:
        den_umap = umap3d(df_well, curves_umap)
    except:
        den_umap = []
    y_pred.loc[ac_umap, ["flag_umap_gen", "flag_umap_ac"]] = 1
    y_pred.loc[acs_umap, ["flag_umap_gen", "flag_umap_acs"]] = 1
    y_pred.loc[den_umap, ["flag_umap_gen", "flag_umap_den"]] = 1
    return y_pred



"""
setup.py

Information used to build the package
"""
from setuptools import find_namespace_packages, setup
import os
import subprocess


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


def get_version():
    ENV = os.environ["ENV"]  # Must be set
    tag = subprocess.check_output(
        ["git", "describe", "--tags", "--exact-match"], encoding="UTF-8"
    ).rstrip()
    if ENV in ["dev", "test"]:
        version = "0." + tag
    elif ENV == "prod":
        version = "1." + tag
    else:
        raise ValueError(f"Unknown environment {ENV}")
    return version


setup(
    name="akerbp.models",
    version=get_version(),
    author="Alfonso M. Canterla",
    author_email="alfonso.canterla@soprasteria.com",
    description="Machine Learning Models for Petrophysics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/akerbp/models/",
    packages=find_namespace_packages(where="src"),
    package_dir={"": "src"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    install_requires=[
        "xgboost==1.3.3",
        "joblib==1.0.1",
        "numpy>=1.19.5",
        "pandas>=1.3.2",
        "scikit-learn>=0.24.2",
    ],
)



from typing import List, Tuple
from pandas.core.frame import DataFrame, Series
import numpy as np


def get_constant_derivatives(
    df: DataFrame, logname: str, th: float = 0.0005
) -> Tuple[List, Series]:
    """
    Returns indices of anomalous samples based on derivatives lower than a threshold,
    and the derivatives

    Args:
        df (DataFrame): dataframe with data to analyse
        logname (str): curve to be flagged, if any
        th (float, optional): threshold to which lower values will be considered
        constant. Defaults to 0.0005.

    Returns:
        Tuple[List, Series]: list if indices of constant derivatives
        and the derivative values
    """
    df["d1"] = np.gradient(df[logname], df["DEPTH"])
    df["d2"] = np.gradient(df["d1"], df["DEPTH"])
    df["derivative"] = np.abs(df["d1"] - df["d2"])
    df["cnst_derivative"] = df["derivative"] < th
    return df[df["cnst_derivative"] == True].index, df["derivative"]


def get_constant_windows(
    df: DataFrame,
    logname: str,
    window_size: int = 10,
    min_periods: int = 5,
    th: float = 0.0005
) -> Tuple[List, Series]:
    """
    Returns indices of anomalous samples based on derivatives lower than a
    threshold within a window, and the derivatives

    Args:
        df (DataFrame): dataframe with data to analyse
        logname (str): curve to be flagged, if any
        window_size (int): size of window. Defaults to 10.
        min_periods (int): min_periods of rolling window. Defaults to 5.
        th (float, optional): threshold to which lower values will be considered
        constant. Defaults to 0.0005.

    Returns:
        Tuple[List, Series]: list if indices of constant derivatives
        and the derivative values
    """
    df["minw"] = df[logname].rolling(
        window_size, min_periods=min_periods, center=True
    ).min()
    df["maxw"] = df[logname].rolling(
        window_size, min_periods=min_periods, center=True
    ).max()
    df["window_derivative"] = np.abs(df["maxw"] - df["minw"])
    df["cnst_window"] = df["window_derivative"] < th
    return df[df["cnst_window"] == True].index, df["window_derivative"]


def get_den_flatlines(df: DataFrame, th: float = 0.002) -> List:
    """
    Returns flatlines specifically for density, where threshold was decided empirically

    Args:
        df (DataFrame): dataframe with data to analyse
        th (float, optional): threshold to which lower values will be considered constant.
        Defaults to 0.002.

    Returns:
        list: indices of anomalous density samples
    """
    if len(df["DEN"]) > 1:
        df["DEN_flat"] = np.gradient(df["DEN"])
        idx = df[
            (np.abs(df["DEN_flat"].diff()) < th) &\
            (np.abs(df["DEN_flat"].diff(-1)) < th)
        ].index
    else:
        idx = []
    return idx


def get_flatlines(
    df: DataFrame,
    logname: str,
    cols: List,
    window: int = 5,
    var_window: int = 10,
    n_cols: int = 2
) -> List:
    """
    Returns indices of anomalous values indicating flatline badlog

    Args:
        df (DataFrame): data to analyze
        logname (str): logname to find anomalies
        cols (list): list of curves to analyse logname against
        window (int, optional): window size for variance calculation.
        Defaults to 5.
        var_window (int, optional): window for rolling values. Defaults to 10.
        n_cols (int, optional): number of columns to compare with. Defaults to 2.

    Returns:
        list: indices of flatlines anomlies
    """
    # get gradients and see if one gradient is not as small as at least three other curves
    cols_grad = [f"{c}_gradient" for c in cols]
    gradient_flag = (
        df[f"{logname}_gradient"] < df[f"{logname}_gradient"].rolling(
            var_window, center=True
        ).std()
    ) & (
        np.sum(df[cols_grad] > df[cols_grad].rolling(
            var_window, center=True
        ).std(), axis=1) > n_cols
    )
    # get variance
    variance = df[[logname] + cols].rolling(window).var()
    variance_flag = (
        variance[logname] < variance[logname].rolling(
            var_window, center=True
        ).std()
    ) & (
        np.sum(variance[cols] > variance[cols].rolling(
            var_window, center=True
        ).std(), axis=1) > n_cols
    )
    # flatlines
    flatlines = gradient_flag | variance_flag
    return flatlines[flatlines == True].index


def find_start_end_indices(df_well: DataFrame) -> Tuple[int, int]:
    """
    Get absolute maximum and minimum indices between AC, ACS and DEN logs

    Args:
        df_well (DataFrame): dataframe with data to analyse

    Returns:
        Tuple[int, int]: min and max indices values
    """
    min_idx = df_well.index.max()
    max_idx = df_well.index.min()
    for col in ["AC", "ACS", "DEN"]:
        min_idx = min(min_idx, df_well[df_well[col].notna()].index.min())
        max_idx = max(max_idx, df_well[df_well[col].notna()].index.max())
    return min_idx, max_idx


def flag_flatline(
    df_well: DataFrame, y_pred: DataFrame = None, **kwargs
) -> DataFrame:
    """
    Returns flatlines indices for AC, ACS and DEN

    Args:
        df_well (DataFrame): data from one well
        y_pred (DataFrame): results. Defaults to None.

    Returns:
        DataFrame: df with added column of results
    """
    print("Method: flatline...")
    expected_curves = kwargs.get(
        "expected_curves",
        set(["DEN", "AC", "ACS", "GR", "NEU", "RDEP", "RMED"])
    )
    cols = kwargs.get("cols", expected_curves)
    ncols = kwargs.get("ncols", 2)

    for col in cols:
        df_well.loc[:, col + "_gradient"] = np.abs(np.gradient(df_well[col]))

    def run_get_flatlines(df_well, cols, ncols, **kwargs):
        window_size = kwargs.get("window", 10)
        var_window_size = kwargs.get("var_window", 20)
        ac_grad_anomalies = get_flatlines(
            df=df_well,
            logname="AC",
            cols=list(set(cols) - set(["AC"])),
            window=window_size,
            var_window=var_window_size,
            n_cols=ncols
        )
        acs_grad_anomalies = get_flatlines(
            df=df_well,
            logname="ACS",
            cols=list(set(cols) - set(["ACS"])),
            window=window_size,
            var_window=var_window_size,
            n_cols=ncols
        )
        den_grad_anomalies = get_flatlines(
            df=df_well,
            logname="DEN",
            cols=list(set(cols) - set(["DEN"])),
            window=window_size,
            var_window=var_window_size,
            n_cols=ncols
        )
        return (
            ac_grad_anomalies.tolist(),
            acs_grad_anomalies.tolist(),
            den_grad_anomalies.tolist()
        )

    ac_grad_anomalies, acs_grad_anomalies, den_grad_anomalies = [], [], []

    ac_flat_grad, ac_derivatives = get_constant_derivatives(df_well, "AC")
    acs_flat_grad, acs_derivatives = get_constant_derivatives(df_well, "ACS")
    den_flat_grad, den_derivatives = get_constant_derivatives(df_well, "DEN")

    ac_grad_anomalies.extend(ac_flat_grad)
    acs_grad_anomalies.extend(acs_flat_grad)
    den_grad_anomalies.extend(den_flat_grad)

    ac_flat_window, ac_window_grad = get_constant_windows(df_well, "AC")
    acs_flat_window, acs_window_grad = get_constant_windows(df_well, "ACS")
    den_flat_window, den_window_grad = get_constant_windows(df_well, "DEN")

    ac_grad_anomalies.extend(ac_flat_window)
    acs_grad_anomalies.extend(acs_flat_window)
    den_grad_anomalies.extend(den_flat_window)

    gen_grad_anomalies = []
    gen_grad_anomalies.extend(ac_grad_anomalies)
    gen_grad_anomalies.extend(acs_grad_anomalies)
    gen_grad_anomalies.extend(den_grad_anomalies)

    # run the function on points where there was constant window or gradient flags
    ac_grad_anomalies_main, acs_grad_anomalies_main, den_grad_anomalies_main =\
        run_get_flatlines(
            df_well.loc[gen_grad_anomalies], cols, ncols, **kwargs
        )

    den_grad_anomalies_main.extend(get_den_flatlines(df_well.loc[gen_grad_anomalies]))
    # pick the start and end of well
    start_end_length = 50
    start_end_window_size = 2
    start_end_var_window_size = 2
    min_idx, max_idx = find_start_end_indices(df_well)
    start_condition = (df_well.index >= min_idx) &\
        (df_well.index <= min_idx + start_end_length)
    end_condition = (df_well.index >= max_idx - start_end_length) &\
        (df_well.index <= max_idx)

    # run the function for end points only with smaller window sizes
    df_well_ends = df_well[start_condition | end_condition]
    ac_grad_anomalies_ends, acs_grad_anomalies_ends, den_grad_anomalies_ends =\
        run_get_flatlines(
            df_well_ends,
            cols,
            ncols,
            **{
                "window": start_end_window_size,
                "var_window": start_end_var_window_size
            }
        )

    ac_grad_anomalies_main.extend(ac_grad_anomalies_ends)
    acs_grad_anomalies_main.extend(acs_grad_anomalies_ends)
    den_grad_anomalies_main.extend(den_grad_anomalies_ends)

    if y_pred is None:
        y_pred = df_well.copy()

    y_pred.loc[:, [
        "flag_flatline_gen",
        "flag_flatline_ac",
        "flag_flatline_acs",
        "flag_flatline_den"
    ]] = 0, 0, 0, 0

    # DEBUG
    flag_columns = ["flag_den_only"]
    lognames = ["ac", "acs", "den"]
    for logname in lognames:
        flag_columns.extend([
            f"flag_main_{logname}",
            f"flag_ends_{logname}",
            f"flag_derivs_{logname}",
            f"flag_windows_{logname}"
        ])
    y_pred.loc[:, flag_columns] = 0

    y_pred.loc[ac_grad_anomalies_main, ["flag_flatline_gen", "flag_flatline_ac"]] = 1
    y_pred.loc[acs_grad_anomalies_main, ["flag_flatline_gen", "flag_flatline_acs"]] = 1
    y_pred.loc[den_grad_anomalies_main, ["flag_flatline_gen", "flag_flatline_den"]] = 1

    return y_pred
